use crate::{
    error::{ChunkingText, Deserializing, Result},
    model::{Definition, Response},
};
use reqwest::Client as HttpClient;
use snafu::ResultExt;
use std::sync::Arc;

pub struct Client(Arc<HttpClient>);

impl Client {
    /// Create a new UrbanDictionary client.
    ///
    /// # Examples
    ///
    /// ```no_run
    /// use reqwest::Client as HttpClient;
    /// use urbandictionary::Client;
    ///
    /// let http_client = HttpClient::new();
    /// let urban_client = Client::new(http_client);
    pub fn new(http_client: impl Into<Arc<HttpClient>>) -> Self {
        Self(http_client.into())
    }

    /// Attempt to retrieve the first definition for a word.
    ///
    /// # Examples
    ///
    /// Retrieve the definition of the word `"cat"`:
    ///
    /// ```no_run
    /// use reqwest::Client as HttpClient;
    /// use urbandictionary::Client;
    ///
    /// # #[tokio::main]
    /// # async fn main() -> Result<(), Box<dyn std::error::Error + Send + Sync>> {
    /// let http_client = HttpClient::new();
    /// let client = Client::new(http_client);
    /// let response = client.define("cat").await?;
    ///
    /// if let Some(definition) = response {
    ///     println!("The definition of cat is: {}", definition.definition);
    /// } else {
    ///     println!("No definition found");
    /// }
    /// # Ok(()) }
    /// ```
    ///
    /// # Errors
    ///
    /// Returns [`Error::Json`] if there was an error deserializing the response
    /// body.
    ///
    /// Returns [`Error::Reqwest`] if there was an error sending the request.
    ///
    /// [`Error::Json`]: ../../enum.Error.html#variant.Json
    /// [`Error::Reqwest`]: ../../enum.Error.html#variant.Reqwest
    pub async fn define(&self, word: &str) -> Result<Option<Definition>> {
        let mut resp = self.definitions(word).await?;

        Ok(if resp.definitions.is_empty() {
            None
        } else {
            Some(resp.definitions.remove(0))
        })
    }

    /// Attempt to retrieve the definitions for a word.
    ///
    /// # Examples
    ///
    /// Retrieve the definitions for the word `"cat"`:
    ///
    /// ```no_run
    /// # #[tokio::main]
    /// # async fn main() -> Result<(), Box<dyn std::error::Error + Send + Sync>> {
    /// use reqwest::Client as HttpClient;
    /// use urbandictionary::Client;
    ///
    /// let http_client = HttpClient::new();
    /// let client = Client::new(http_client);
    ///
    /// let response = client.definitions("cat").await?;
    ///
    /// if let Some(definition) = response.definitions.first() {
    ///     println!(
    ///         "The first definition of cat is: {}",
    ///         definition.definition,
    ///     );
    /// } else {
    ///     println!("No definitions found");
    /// }
    /// # Ok(()) }
    /// ```
    ///
    /// # Errors
    ///
    /// Returns [`Error::Json`] if there was an error deserializing the response
    /// body.
    ///
    /// Returns [`Error::Reqwest`] if there was an error sending the request.
    ///
    /// [`Error::Json`]: ../../enum.Error.html#variant.Json
    /// [`Error::Reqwest`]: ../../enum.Error.html#variant.Reqwest
    pub async fn definitions(&self, word: &str) -> Result<Response> {
        let uri = format!("http://api.urbandictionary.com/v0/define?term={}", word);

        let text = self
            .0
            .get(&uri)
            .send()
            .await?
            .text()
            .await
            .context(ChunkingText)?;

        serde_json::from_str(&text).with_context(|| Deserializing { text })
    }
}
